#!/usr/bin/perl

use strict;
use warnings;

my $FORMAT_VERSION = '2012_01_02_00';
my @version_parts = split(/_/, $FORMAT_VERSION);

my $version_string = pack("nCCC", int($version_parts[0]),
		int($version_parts[1]),
		int($version_parts[2]),
		int($version_parts[3]));

my $gamefile;
my $name;

sub fail {
	print STDERR "Error: $_[0]\n" if(@_ && length($_[0]));
	exit 1;
}

sub put_padded_string{
	my ($text, $width) = @_;
	return $text . "\x00"x($width-length($text));
}

sub collect_tilesets {
	opendir(TILEROOT, "tilesets") || die "Could not open tileset root: $!";
	my @tiledirs = readdir TILEROOT; #XXX: Assumes sorted
	closedir TILEROOT;
	foreach my $tiledir (@tiledirs){
		next if($tiledir=~/^\./ || !-d "tilesets/$tiledir");
		print $gamefile 'T';
		open(TILESET, '<', "tilesets/$tiledir/tileset.c2t") || die "Could not open tileset file: $!";
		
		my $state = 0;
		my $j=0;
		foreach my $line (<TILESET>){
			chomp $line;
			$line =~ s/^\s*//;
			next if(!length($line) || $line=~/^#/);
			if($state == 0){
				if($line=~/^name=([a-zA-Z0-9_\- ]{1,16})$/){
					print "Tileset name: $1\n";
					print $gamefile put_padded_string($1, 16);
					$state++;
				}else{
					fail "Syntax error in tileset '$tiledir': '$line'";
				}
			}elsif($state == 1){
				if($line=~/^\[([0-9]{2})\]$/ && $1 == $j){
					$state++;
					print $gamefile "t";
				}else{
					fail "Syntax error in tileset '$tiledir': '$line'";
				}
			}elsif($state == 2){
				if($line=~/^name=([a-zA-Z0-9_\- ]{1,16})$/){
					print "Tile name: $1\n";
					print $gamefile put_padded_string($1, 16);
					$state++;
				}else{
					fail "Syntax error in tileset '$tiledir': '$line'";
				}
			}elsif($state == 3){
				if($line=~/^passable=(true|false)$/){
					print "Passable = $1\n";
					print $gamefile ($1 eq 'true'?"\x01":"\x00");
					$state++;
				}else{
					fail "Syntax error in tileset '$tiledir': '$line'";
				}
			}elsif($state == 4){
				if($line=~/^deadly=(true|false)$/){
					print "Deadly = $1\n\n";
					print $gamefile ($1 eq 'true'?"\x01":"\x00");
					$state = 1;
				}else{
					fail "Syntax error in tileset '$tiledir': '$line'";
				}
				open(TILEPNM, "tilesets/$tiledir/tile_" . ($j<10?"0$j":"$j") . ".pnm") || die "Could not open tile PNM: $!";
				binmode(TILEPNM);
				seek(TILEPNM, 0, 2);
				my $tilesize = tell(TILEPNM);
				if($tilesize > 0xFFFF){
					fail "Tile image too large";
				}
				print $gamefile pack("n", $tilesize);
				seek(TILEPNM, 0, 0);
				while(<TILEPNM>){
					print $gamefile $_;
				}
				close TILEPNM;
				$j++;
			}
		}
		print $gamefile "\xFF";
		fail "Syntax error in tileset '$tiledir'" if($state != 1);
		close TILESET;
	}
}

sub collect_entities {
	opendir(ENTITYROOT, "entities") || die "Could not open entity root: $!";
	my @entitydirs = readdir ENTITYROOT;
	closedir ENTITYROOT;
	my $j=0;
	foreach my $entitydir (@entitydirs){
		next if($entitydir=~/^\./ || !-d "entities/$entitydir");
		print $gamefile 'E';
		open(ENTITY, '<', "entities/$entitydir/entity.c2e") || die "Could not open entity file: $!";
		
		my $state = 0;
		foreach my $line (<ENTITY>){
			chomp $line;
			$line =~ s/^\s*//;
			next if(!length($line) || $line=~/^#/);
			if($state == 0){
				if($line=~/^name=([a-zA-Z0-9_\- ]{1,16})$/){
					print "Entity name: $1\n";
					print $gamefile put_padded_string($1, 16);
					$state++;
				}else{
					fail "Syntax error in entity '$entitydir': '$line'";
				}
			}elsif($state == 1){
				if($line=~/^width=([0-9]{1,3})$/){
					print "Width: $1\n";
					print $gamefile pack("C", $1);
					$state++;
				}else{
					fail "Syntax error in entity '$entitydir': '$line'";
				}
			}elsif($state == 2){
				if($line=~/^height=([0-9]{1,3})$/){
					print "Height: $1\n";
					print $gamefile pack("C", $1);
					$state++;
				}else{
					fail "Syntax error in entity '$entitydir': '$line'";
				}
			}elsif($state == 3){
				if($line=~/^hitbox_left=([0-9]{1,3})$/){
					print "Hitbox offset left: $1\n";
					print $gamefile pack("C", $1);
					$state++;
				}else{
					fail "Syntax error in entity '$entitydir': '$line'";
				}
			}elsif($state == 4){
				if($line=~/^hitbox_top=([0-9]{1,3})$/){
					print "Hitbox offset top: $1\n";
					print $gamefile pack("C", $1);
					$state++;
				}else{
					fail "Syntax error in entity '$entitydir': '$line'";
				}
			}elsif($state == 5){
				if($line=~/^hitbox_right=([0-9]{1,3})$/){
					print "Hitbox offset right: $1\n";
					print $gamefile pack("C", $1);
					$state++;
				}else{
					fail "Syntax error in entity '$entitydir': '$line'";
				}
			}elsif($state == 6){
				if($line=~/^hitbox_bottom=([0-9]{1,3})$/){
					print "Hitbox offset bottom: $1\n\n";
					print $gamefile pack("C", $1);
					$state++;
				}else{
					fail "Syntax error in entity '$entitydir': '$line'";
				}
				open(ENTITYPNM, "entities/$entitydir/entity.pnm") || die "Could not open entity PNM: $!";
				binmode(ENTITYPNM);
				seek(ENTITYPNM, 0, 2);
				my $entitysize = tell(ENTITYPNM);
				if($entitysize > 0xFFFF){
					fail "Entity image too large";
				}
				print $gamefile pack("n", $entitysize);
				seek(ENTITYPNM, 0, 0);
				while(<ENTITYPNM>){
					print $gamefile $_;
				}
				close ENTITYPNM;
				$j++;
				$state = -1;
			}else{
				fail "Oops'ed in entity '$entitydir'" if($state != -1);
			}
		}
		print $gamefile "\xFF";
		fail "Error in entity '$entitydir' (state:$state)" if($state != -1);
		close ENTITY;
	}
}

sub collect_levels {
	opendir(LEVELROOT, "levels") || die "Could not open level root: $!";
	my @leveldirs = readdir LEVELROOT;
	closedir LEVELROOT;
	foreach my $levelfile (@leveldirs){
		next if($levelfile=~/^\./ || !-f "levels/$levelfile" || $levelfile!~/.c2l$/);
		print $gamefile 'L';
		open(LEVEL, '<', "levels/$levelfile") || die "Could not open level file: $!";
		
		my $state = 0;
		my $level_length = 0;
		my $j=0;
		foreach my $line (<LEVEL>){
			chomp $line;
			$line =~ s/^\s*//;
			next if(!length($line) || $line=~/^#/);
			if($state == 0){
				if($line=~/^name=([a-zA-Z0-9_\- ]{1,16})$/){
					print "Level name: $1\n";
					print $gamefile put_padded_string($1, 16);
					$state++;
				}else{
					fail "Syntax error in level '$levelfile': '$line'";
				}
			}elsif($state == 1){
				if($line=~/^start_x=([0-9]{1})$/){
					print "Start x: $1\n";
					print $gamefile pack("C", $1);
					$state++;
				}else{
					fail "Syntax error in level '$levelfile': '$line'";
				}
			}elsif($state == 2){
				if($line=~/^start_y=([0-9]{1})$/){
					print "Start y: $1\n";
					print $gamefile pack("C", $1);
					$state++;
				}else{
					fail "Syntax error in level '$levelfile': '$line'";
				}
			}elsif($state == 3){
				if($line=~/^c=([a-zA-Z ]+)$/){
					my $row = $1;
					if($j == 0){
						$level_length = length($row);
						print "Width: $level_length\n\n";
						print $gamefile pack("n", length($row));
					}else{
						if(length($row) != $level_length){
							fail "Level length not uniform (".length($row).", while level length is set to $level_length)";
						}
					}
					#my $or = $row;
					#$row =~ tr/ a-zA-Z/0-52/;
					#print "CHANGED $or -> $row\n";
					my %num = ();
					my $k=0;
					$num{' '} = $k++;
					foreach ('a'..'z'){
						$num{$_} = $k++;
					}
					foreach ('A'..'Z'){
						$num{$_} = $k++;
					}
					my @chars = split(//, $row);
					foreach(@chars){
						$_ = $num{$_};
						if($_ >= 26){
							$_ -= 26;
							$_ |= 0b10000000;
						}
						$_ = pack("C", $_);
						print $gamefile $_;
					}
					
					$state++ if($j == 9);
					$j++;
				}else{
					fail "Syntax error in level '$levelfile'";
				}
			}else{
				fail "Oops'ed in level '$levelfile'" if($state != -1);
			}
		}
		print $gamefile "\xFF";
		#fail "Error in level '$levelfile' (state:$state)" if($state != -1);
		close ENTITY;
	}
}

if(@ARGV != 1){
	print STDERR "Usage: compile_gamefile.pl <gamefile>\n";
	fail;
}

print <<EOF;
compile_gamefile.pl
Copyright (c) 2011, Chris Johan Smeele (cjsx9g)
All rights reserved.

EOF

my $gamedescription_filename = $ARGV[0];

if(!-f $gamedescription_filename || !-r $gamedescription_filename){
	fail "Game description file '$gamedescription_filename' does not exist or isn't readable.\n";
	exit 1;
}

print "Output file version: $FORMAT_VERSION\n";

my $lives = 0;

my $gamedescription_file;
open($gamedescription_file, '<', $gamedescription_filename) || fail "Could not open game description file for reading: $!";
my $i=0;
while(<$gamedescription_file>){
	$i++;
	chomp;
	s/^\s*//;
	#s/\s*$//;
	next if(!length($_) || /^#/);
	if(/^name=([a-zA-Z0-9_\- ]{1,16})$/){
		print "Game name: '$1'\n";
		$name = $1;
	}elsif(/^lives=([0-9]{1})$/){
		print "Lives: $1\n\n";
		$lives = $1;
	}else{
		fail "Syntax error at $gamedescription_filename, line $i.\n";
	}
}
close $gamedescription_file;

if(!defined $name || !length($name)){
	fail "No (valid) name specified in game description file";
}

my $gamefilename = "$name";
$gamefilename =~ s/ /_/g;
$gamefilename .= '.c2d';

open($gamefile, '>', $gamefilename) || fail "Could not open game file for writing: $!";
binmode($gamefile);
print $gamefile "C2D\x05\x39" . $version_string . "\x00";
print $gamefile put_padded_string($name, 16);
print $gamefile pack("C", $lives);

print "--------\nTilesets\n\n";
collect_tilesets;
print "--------\nEntities\n\n";
collect_entities;
print "--------\nLevels\n\n";
collect_levels;

print $gamefile "\xff";


close $gamefile;








